option explicit

''
'' FreeBASIC Sokoban
'' (C)2005 by Peter Bakota <bakota@gmail.com>
''
'' This program is free software; you can redistribute it and/or modify it
'' but WITHOUT ANY WARRANTY; without even the implied warranty of
'' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
''

declare sub draw_level(byval level_num as integer, byref startx, starty as integer)
declare sub draw_object(byval x, y as integer, byval obj as byte)

' Game block graphics size
#define BLK_W   16
#define BLK_H   16

' Game state
#define S_NONE          0
#define S_ABORT         1
#define S_SKIP_LEVEL    2
#define S_BACK_LEVEL    3
#define S_LEVEL_DONE    4
#define S_UNDO_LEVEL    5

' Game objects
#define T_EMPTY     0
#define T_BOX       1
#define T_WALL      2
#define T_PLAYER1   3
#define T_PLAYER2   4

' Target
#define F_TARGET    64
#define M_TARGET    63

' Helper macros
#define IS_TARGET(x) ((x) and F_TARGET)<>0
#define M_OBJ(x) (x and M_TARGET)
#define S_OBJ(x, a) x = (x and F_TARGET) or a

' Moves for take back
#define PUSH_NONE   0
#define PUSH_LEFT   1
#define PUSH_RIGHT  2
#define PUSH_UP     3
#define PUSH_DOWN   4

' 640x480x8
screen 18
windowtitle "FreeBASIC - Sokoban"

title:
data " ###  ##  #  #  ##  ###   ##  #  #"
data "#$   #$ # # #  #  # #  # # $# ## #"
data " ##  #$$# ##   #  # ###  #### #$##"
data "  $# # $# #$#  #$ # #$ # #$ # # $#"
data "###   ##  # $#  ##  ###  #  # # $#"

' Level set
dim shared levels_v() as string
' Max level in set
dim shared max_level as integer
' current level map
dim shared map_v() as byte
' Max targets in level
dim shared max_targets as integer
' Playfield offsets (centering on screen)
dim shared plf_x as integer
dim shared plf_y as integer
' level set name
dim shared current_set as string

'===========================================
sub wait_key
'===========================================
    ' Clear input buffer
    do: loop while inkey$<>""
    ' Wait for key press
    do: loop until inkey$<>""
end sub

'===========================================
sub add_level(byval level as string)
'===========================================
    redim preserve levels_v(max_level+1)
    levels_v(max_level) = level
    max_level += 1
end sub

'===========================================
sub load_set(byval set_name as string)
'===========================================
    dim s as string
    dim level as string
    dim cwd as string = "levels/"
    dim c as string
    dim eol as integer
    dim f as integer

    redim levels_v(0) '' clear levels
    level = ""
    max_level = 0
    
    f = freefile

    open cwd + set_name + ".xsb" for input as #f
    do
        s = ""
        eol = 0
        do
            c = input$(1, #f)
            if c<>chr$(10) and c<>chr$(13) and c<>"" then s += c else eol = 1 
        loop until eol=1

        if len(s)>0 then
            ' Every valid line must begin with " " or with "#" and must
            ' have min. one "#"
            if left$(s, 1)<>" " and left$(s, 1)<>"#" or instr(s, "#")=0 then
                s = ""
            end if
        end if

        if s<>"" then
            level += s + chr$(13)
        else
            if len(level)>0 then
                add_level level
                level = ""
            end if
        end if

    loop until eof(f)
    close #f

    if len(level)>0 then add_level level

    current_set = set_name

end sub

'===========================================
function title_screen as integer
'===========================================
    ' Option lines
    dim opt_x(3) as integer = { 19, 20, 21, 23 }
    dim i, j as integer
    dim k as string

    color 67,0
    cls

    ' Reset playfield offsets
    plf_x = 0
    plf_y = 0

    ' Random background graphics
    dim x, y, c as integer
    dim o(2) as integer = { T_BOX, T_WALL }
    for i=1 to 350
        x = rnd*40
        y = rnd*25
        c = rnd*1
        draw_object x, y, o(c)
    next

    ' Sokoban text border
    line (20, 40) - (620, 200), 66, bf
    line (20, 40) - (620, 200), 44, b

    ' Draw big 'SOKOBAN' text
    color 15,66
    locate 5, 7: print "FreeBASIC"
    x = 3
    y = 5
    restore title
    for i=1 to 5
        read k
        for j=1 to len(k)
            select case mid$(k, j, 1)
            case "#"
                draw_object x, y, T_WALL
            case "$"
                draw_object x, y, T_BOX
            end select
            x += 1
        next
        x = 3
        y += 1
    next

    ' Credits
    color 15, 0
    locate 29, 28: print "http://www.freebasic.net"
    
    ' Menu border & background
    line (354, 284) - (472, 384), 25, bf
    line (350, 280) - (468, 380), 66, bf
    line (350, 280) - (468, 380), 44, b
    ' Draw menu
    color 15, 66
    locate opt_x(0), 48: print "Start";
    locate opt_x(1), 48: print "Level Set";
    locate opt_x(2), 48: print "Help";

    locate opt_x(3), 48: print "Quit";

    ' Current choice
    dim choice as integer = 0

    ' Menu select loop
    do

        ' Erase menu pointer
        locate opt_x(choice), 46: print " ";
        if k=chr$(255)+chr$(72) then
            if choice>0 then choice -= 1 else choice = 3
        else 
            if k=chr$(255)+chr$(80) then
                if choice<3 then choice += 1 else choice = 0
            end if
        end if
        ' Draw menu pointer
        locate opt_x(choice), 46: print ">";

        ' Wait for action ...
        do: k = inkey$: loop until k<>""

    loop until k=chr$(13)

    return choice

end function

'===========================================
sub draw_object(byval x, y as integer, byval obj as byte)
'===========================================
    dim sx as integer = x*BLK_W+plf_x
    dim sy as integer = y*BLK_H+plf_y

    ' Erase object background
    color 0
    line (sx, sy) - (sx+BLK_W-2, sy+BLK_H-2),,bf

    select case M_OBJ(obj)
    case T_BOX
        ' Box
        color 4
        line (sx, sy) - (sx+BLK_W-2, sy+BLK_H-2),,bf
        color 42
        line (sx, sy) - (sx+BLK_W-2, sy+BLK_H-2),,b
        line (sx, sy) - (sx+BLK_W-2, sy+BLK_H-2)
        line (sx+BLK_W-2, sy) - (sx, sy+BLK_H-2)
    case T_WALL
        ' Wall
        line (sx, sy) - (sx+BLK_W-2, sy+BLK_H-2),27,bf
        line (sx, sy) - (sx+BLK_W-2, sy),30
        line (sx+BLK_W-2, sy) - (sx+BLK_W-2, sy+BLK_H-2),30
        line (sx, sy) - (sx, sy+BLK_H-2),26
        line (sx, sy+BLK_H-2) - (sx+BLK_W-2, sy+BLK_H-2),26
    case T_PLAYER1, T_PLAYER2
        ' Player
        color 44
        circle (sx+BLK_W/2, sy+BLK_W/2),BLK_W/2-2,,,,,F
        if obj=T_PLAYER2 then
            color 0
            circle (sx+BLK_W/4+2, sy+BLK_W/4+2),BLK_W/10,,,,,F
            circle (sx+BLK_W/4+BLK_W/4+3, sy+BLK_W/4+2),BLK_W/10,,,,,F
        end if
    case T_EMPTY
        ' Empty or target
        if IS_TARGET(obj) then
            color 40
            circle (sx+BLK_W/2, sy+BLK_H/2),2,,,,,F
        end if
    end select
end sub

'===========================================
sub draw_level(byval level_num as integer, byref startx, starty as integer)
'===========================================
    ' Get level data
    dim level as string = levels_v(level_num)
    dim i as integer
    dim x, y as integer
    ' Level max. width and height
    dim max_x as integer = 0
    dim max_y as integer = 0

    ' Clear screen
    color 15,0
    cls
    
    ' Clear level map
    redim map_v(40,32)
    
    ' No targets yet!
    max_targets = 0

    x = 0
    y = 0
    ' prepare level data
    for i=1 to len(level)
        select case mid$(level, i, 1)
        case "#"
            ' Wall
            map_v(x, y) = T_WALL: x += 1
        case "."
            ' Target
            map_v(x, y) = F_TARGET + T_EMPTY: x += 1
        case "$"
            ' Box
            map_v(x, y) = T_BOX: x += 1
        case " "
            ' Empty
            map_v(x, y) = T_EMPTY: x += 1
        case "*"
            ' Target with box
            map_v(x, y) = T_BOX + F_TARGET: x += 1
        case chr$(13)
            ' Next line
            y += 1
            x = 0
        case "@", "+"
            if mid$(level, i, 1)="+" then
                ' Player on target
                map_v(x, y) = F_TARGET + T_EMPTY
            end if
            ' Player start pos
            startx = x
            starty = y
            x += 1
        end select
        if x*BLK_W>max_x then max_x = x*BLK_W
        if y*BLK_H>max_y then max_y = y*BLK_H
    next

    ' Calculate playfield offsets
    plf_x = (640 - max_x) / 2
    plf_y = (480 - max_y) / 2

    ' Draw map graphics and count targets
    for x=0 to 39
        for y=0 to 24
            draw_object x, y, map_v(x,y)
            if IS_TARGET(map_v(x,y)) then max_targets += 1
        next
    next

    color 44
    locate 3, 31: print using "Level ### of ###";level_num+1;max_level;
    dim k as string = "("+current_set+")"
    color 7
    locate 2, (80-len(k))/2: print k;
end sub

'===========================================
sub show_stats(byval total_moves as integer)
'===========================================
    color 15

    locate 28, 26: print using "Level solved with #### moves.";total_moves;
    locate 29, 16: print "Press 'ENTER' key when ready for the next level..."

    do while inkey$<>chr$(13): loop
end sub

'===========================================
sub game_loop
'===========================================
    ' Level counter
    dim current_level as integer = 0
    ' Player positions
    dim playerx, playery as integer
    dim k as string
    dim i as integer
    ' Player background
    dim player_old as byte = T_EMPTY
    ' Game state
    dim state as integer
    ' Take back vars
    dim last_push as integer = PUSH_NONE
    dim last_x, last_y as integer
    ' Move counter
    dim total_moves as integer

    do
        ' Draw level
        draw_level current_level, playerx, playery

        ' Initialize game state
        state = S_NONE
        total_moves = 0
        
        ' Clear keyboard buffer
        do: k = inkey$: loop while k<>""

        do
            ' restore player background eg. remove player graphics from screen
            draw_object playerx, playery, player_old

            ' Handle user key actions
            select case left$(k,1)
            case chr$(27)
                ' Esc = abort
                state = S_ABORT
            case chr$(255)
                ' Special keys
                select case mid$(k, 2, 1)
                case chr$(72)
                    ' up
                    if M_OBJ(map_v(playerx, playery-1))=T_BOX and M_OBJ(map_v(playerx,playery-2))=T_EMPTY then
                        S_OBJ(map_v(playerx, playery-2), T_BOX)
                        draw_object playerx, playery-2, T_BOX
                        S_OBJ(map_v(playerx, playery-1), T_EMPTY)
                        playery -= 1
                        last_push = PUSH_UP
                        last_x = playerx
                        last_y = playery
                        total_moves += 1 ' Count moves
                    else
                        if M_OBJ(map_v(playerx, playery-1))=T_EMPTY then 
                            playery -= 1
                            total_moves += 1 ' Count moves
                        end if
                    end if
                case chr$(80)
                    ' down
                    if M_OBJ(map_v(playerx, playery+1))=T_BOX and M_OBJ(map_v(playerx,playery+2))=T_EMPTY then
                        S_OBJ(map_v(playerx, playery+2), T_BOX)
                        draw_object playerx, playery+2, T_BOX
                        S_OBJ(map_v(playerx, playery+1), T_EMPTY)
                        playery += 1
                        last_push = PUSH_DOWN
                        last_x = playerx
                        last_y = playery
                        total_moves += 1 ' Count moves
                    else
                        if M_OBJ(map_v(playerx, playery+1))=T_EMPTY then 
                            playery += 1
                            total_moves += 1 ' Count moves
                        end if
                    end if
                case chr$(75)
                    ' left
                    if M_OBJ(map_v(playerx-1, playery))=T_BOX and M_OBJ(map_v(playerx-2,playery))=T_EMPTY then
                        S_OBJ(map_v(playerx-2, playery), T_BOX)
                        draw_object playerx-2, playery, T_BOX
                        S_OBJ(map_v(playerx-1, playery), T_EMPTY)
                        playerx -= 1
                        last_push = PUSH_LEFT
                        last_x = playerx
                        last_y = playery
                        total_moves += 1 ' Count moves
                    else
                        if M_OBJ(map_v(playerx-1, playery))=T_EMPTY then
                            playerx -= 1
                            total_moves += 1 ' Count moves
                        end if
                    end if
                case chr$(77)
                    ' right
                    if M_OBJ(map_v(playerx+1, playery))=T_BOX and M_OBJ(map_v(playerx+2,playery))=T_EMPTY then
                        S_OBJ(map_v(playerx+2, playery), T_BOX)
                        draw_object playerx+2, playery, T_BOX
                        S_OBJ(map_v(playerx+1, playery), T_EMPTY)
                        playerx += 1
                        last_push = PUSH_RIGHT
                        last_x = playerx
                        last_y = playery
                        total_moves += 1 ' Count moves
                    else
                        if M_OBJ(map_v(playerx+1, playery))=T_EMPTY then
                            playerx += 1
                            total_moves += 1 ' Count moves
                        end if
                    end if
                case chr$(81) ' Pg down
                    ' skip level
                    state = S_SKIP_LEVEL
                case chr$(73) ' Pg up
                    ' back level
                    state = S_BACK_LEVEL
                end select
            case chr$(8) ' Backspace
                ' Undo move
                select case last_push
                case PUSH_UP
                    playerx = last_x
                    playery = last_y
                    ' Reverse push up / pull down
                    playery += 1
                    S_OBJ(map_v(playerx, playery-1), T_BOX)
                    draw_object playerx, playery-1, T_BOX
                    S_OBJ(map_v(playerx, playery-2), T_EMPTY)
                    draw_object playerx, playery-2, map_v(playerx, playery-2)
                case PUSH_DOWN
                    playerx = last_x
                    playery = last_y
                    ' Reverse push down / pull up
                    playery -= 1
                    S_OBJ(map_v(playerx, playery+1), T_BOX)
                    draw_object playerx, playery+1, T_BOX
                    S_OBJ(map_v(playerx, playery+2), T_EMPTY)                    
                    draw_object playerx, playery+2, map_v(playerx, playery+2)
                case PUSH_RIGHT
                    playerx = last_x
                    playery = last_y
                    ' Reverse push right / pull left
                    playerx -= 1
                    S_OBJ(map_v(playerx+1, playery), T_BOX)
                    draw_object playerx+1, playery, T_BOX
                    S_OBJ(map_v(playerx+2, playery), T_EMPTY)
                    draw_object playerx+2, playery, map_v(playerx+2, playery)
                case PUSH_LEFT
                    playerx = last_x
                    playery = last_y
                    ' Reverse push left / pull right
                    playerx += 1
                    S_OBJ(map_v(playerx-1, playery), T_BOX)
                    draw_object playerx-1, playery, T_BOX
                    S_OBJ(map_v(playerx-2, playery), T_EMPTY)
                    draw_object playerx-2, playery, map_v(playerx-2, playery)
                end select
                last_push = PUSH_NONE
            case "r","R"
                ' Undo level
                state = S_UNDO_LEVEL
            end select

            ' Draw player graphics
            player_old = map_v(playerx, playery)
            draw_object playerx, playery, T_PLAYER2

            if state=S_NONE then
                ' Check is level done
                dim x, y, target_ok as integer
                target_ok = 0
                for x=0 to 39
                    for y=0 to 24
                        if IS_TARGET(map_v(x, y)) and M_OBJ(map_v(x, y))=T_BOX then target_ok += 1
                    next
                next
                if target_ok=max_targets then state = S_LEVEL_DONE

                ' Wait for user action
                if state=S_NONE then do : k = inkey$: loop until k<>""
            end if

        loop until state<>S_NONE

        ' Game state
        select case state
        case S_ABORT
            exit do
        case S_LEVEL_DONE
            ' Show statistics
            show_stats total_moves
            current_level += 1
            if current_level = max_level then current_level = 0
        case S_SKIP_LEVEL
            ' Skip level
            current_level += 1
            if current_level >= max_level then current_level = max_level-1
        case S_BACK_LEVEL
            ' Back level
            if current_level>0 then current_level -= 1
        case S_UNDO_LEVEL
            ' Undo level
        end select

    loop

end sub

'===========================================
sub show_level_sets_dir
'===========================================
#define MAX_NAME_LEN 12
#define MAX_LEVELS 100

    dim ls as string
    dim x, y as integer
    dim num as integer = 1

    color 15,0
    cls
    
    ls = dir$("levels/*.xsb")

    locate 2, 28: print "Select level set to play"
    x = 4
    y = 5
    do while len(ls)>0 and num<MAX_LEVELS ' Limit max levels
        ls = left$(ls, len(ls)-4) ' Strip extension
        
        ' Save set name
        redim preserve set_name(num+1) as string
        set_name(num) = ls

        ' make name size exactly MAX_NAME_LEN length
        if len(ls)>MAX_NAME_LEN-1 then ls = left$(ls, MAX_NAME_LEN-4)+"..."
        if len(ls)<MAX_NAME_LEN-1 then ls = ls + string$(MAX_NAME_LEN-len(ls), " ")
        
        color 44
        ' Set id
        locate y, x: print using "## ";num;
        color 7
        ' Set name (stripped!)
        print ls;
        num += 1
        x += MAX_NAME_LEN+2
        if x+MAX_NAME_LEN>=80 then
            x = 4
            y += 1
        end if
        ls = dir$
    loop

    dim k as integer
    dim set_select as integer = 0

    do
        color 15
        locate 28, 28: print "Enter set number: ";: input k
        if k<num then
            set_select = k
        else
            locate 28, 28: print string$(40, " ");
            color 40
            locate 29, 23: print "Wrong number, please try again..."
        end if
    loop until k=0 or set_select<>0

    if set_select<>0 then
        load_set set_name(set_select)
    end if

end sub

'===========================================
sub show_help
'===========================================
    color 15,0
    cls
    
    locate 9, 15: print "Keyboard help"
    locate 11, 12: color 44: print "Cursor left";: color 7: print tab(30);"move/push to left";
    locate 12, 12: color 44: print "Cursor right";: color 7: print tab(30);"move/push to right";
    locate 13, 12: color 44: print "Cursor up";: color 7: print tab(30);"move/push to up";
    locate 14, 12: color 44: print "Cursor down";: color 7: print tab(30);"move/push to down";
    locate 15, 12: color 44: print "Backspace";: color 7: print tab(30);"Undo last move";
    locate 16, 12: color 44: print "R";: color 7: print tab(30);"Undo level";
    locate 17, 12: color 44: print "PgDown";: color 7: print tab(30);"Go to next level";
    locate 18, 12: color 44: print "PgUp";: color 7: print tab(30);"Go to previous level";
    locate 19, 12: color 44: print "Esc";: color 7: print tab(30);"Back to main menu";
    
    color 15
    locate 28, 32: print "Press any key..."
    
    wait_key
end sub


'===========================================
' MAIN
'===========================================

color 15,0
cls

locate 15, 36: print "SOKOBAN!"

' load default set
load_set "default"

' main loop
do
    select case title_screen
    case 0
        ' start
        game_loop
    case 1
        ' level sets
        show_level_sets_dir
    case 2
        ' Show help
        show_help
    case 3
        exit do
    end select
loop

'' ESC = exit
''do: loop until inkey$=chr$(27)

end
